import logging
import os
import sys
from io import TextIOWrapper

import anyio

from minimcp.server import MiniMCP

from .math_mcp import math_mcp

# Configure logging globally for the demo server
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[
        logging.FileHandler(os.environ.get("MCP_SERVER_LOG_FILE", "mcp_server.log")),
        logging.StreamHandler(),  # Also log to stderr
    ],
)
logger = logging.getLogger(__name__)


async def stdio_transport(mcp: MiniMCP):
    """
    stdio_transport interfaces MiniMCP with stdio by reading from
    the current process stdin and writing to stdout.
    """

    stdin = anyio.wrap_file(TextIOWrapper(sys.stdin.buffer, encoding="utf-8"))
    stdout = anyio.wrap_file(TextIOWrapper(sys.stdout.buffer, encoding="utf-8"))

    async def handle_message(line: str):
        logger.info("Handling incoming message: %s", line)

        response = await mcp.handles(line)
        if response:
            logger.info("Writing response message to stdio: %s", response)
            await stdout.write(response + "\n")
            await stdout.flush()

    logger.info("MiniMCP: Started stdio server, listening for messages...")
    try:
        # stdio is expected to connect the server with one single client,
        # hence not deploying advanced concurrency management or handling backpressure
        async with anyio.create_task_group() as tg:
            async for line in stdin:
                tg.start_soon(handle_message, line)
    except (KeyboardInterrupt, anyio.get_cancelled_exc_class()):
        print("\nCtrl+C detected, exiting gracefully...")


def main():
    anyio.run(stdio_transport, math_mcp)


if __name__ == "__main__":
    main()



import json
import logging
from typing import Any, Generic

import anyio
import mcp.shared.version as version
import mcp.types as types
from mcp.server.lowlevel.server import NotificationOptions, Server
from mcp.shared.exceptions import McpError
from pydantic import ValidationError

import minimcp.server.json_rpc as json_rpc
from minimcp.server.managers.context_manager import ContextManager, ScopeT
from minimcp.server.utils import to_dict

from .exceptions import UnsupportedRPCMessageType
from .managers.tool_manager import ToolManager

logger = logging.getLogger(__name__)


class MiniMCP(Generic[ScopeT]):
    _core: Server
    _notification_options: NotificationOptions | None = None

    _timeout: int
    _max_concurrency: int
    _raise_exceptions: bool

    tool: ToolManager
    context: ContextManager[ScopeT]

    def __init__(
        self,
        name: str,
        version: str | None = None,
        instructions: str | None = None,
        timeout: int = 30,
        max_concurrency: int = 100,
        raise_exceptions: bool = False,
    ) -> None:
        """
        Initialize the MCP server.

        Args:
            name: The name of the MCP server.
            version: The version of the MCP server.
            instructions: The instructions for the MCP server.

            timeout: Time in seconds after which a message handler will timeout.
            max_concurrency: The maximum number of message handlers that could be run at
                the same time, beyond which the handle() calls will be blocked.
            raise_exceptions: Whether to raise uncaught exceptions while handling messages.
        """
        self._timeout = timeout
        self._max_concurrency = max_concurrency

        self._raise_exceptions = raise_exceptions
        self._limiter = anyio.CapacityLimiter(self._max_concurrency)

        # TODO: Add support for automatic server-to-client notifications
        self._notification_options = NotificationOptions(
            prompts_changed=False,
            resources_changed=False,
            tools_changed=False,
        )

        # Setup core
        self._core = Server(name=name, version=version, instructions=instructions)
        self._core.request_handlers[types.InitializeRequest] = self._initialize_handler

        # Setup managers
        self.tool = ToolManager(self._core)
        self.context = ContextManager()

    # --- Properties ---
    @property
    def name(self) -> str:
        return self._core.name

    @property
    def instructions(self) -> str | None:
        return self._core.instructions

    @property
    def version(self) -> str | None:
        return self._core.version

    # --- Handlers ---
    async def handle(self, message: dict[str, Any], scope: ScopeT | None = None) -> dict[str, Any] | None:
        try:
            rpc_msg = types.JSONRPCMessage.model_validate(message)

            async with self._limiter:
                with anyio.fail_after(self._timeout):
                    with self.context.active(rpc_msg, scope):
                        response = await self._handle_rpc_msg(rpc_msg)

        # --- Centralized MCP error handling - All expected exceptions must be handled here ---
        except ValidationError as e:
            logger.error("Invalid Message: %s", e)
            response = json_rpc.build_error_message(types.INVALID_REQUEST, message, e)
        except UnsupportedRPCMessageType as e:
            logger.error("Unsupported Message Type: %s", e)
            response = json_rpc.build_error_message(types.INVALID_REQUEST, message, e)
        except McpError as e:
            logger.error("Error While Handling Message: %s", e)
            response = json_rpc.build_error_message(types.INTERNAL_ERROR, message, e, e.error)
        except TimeoutError as e:
            logger.error("Message Handler Timed Out: %s", message)
            response = json_rpc.build_error_message(types.INTERNAL_ERROR, message, e)
        except anyio.get_cancelled_exc_class():
            logger.info("Message Cancelled: %s", message)
            response = None
        except Exception as e:
            logger.error("Uncaught Exception: %s", e)
            if self._raise_exceptions:
                raise
            response = json_rpc.build_error_message(types.INTERNAL_ERROR, message, e)

        if response is None:
            logger.info(f"No response returned for message: {message}")
            return None

        return to_dict(response)

    async def handles(
        self, message: str, scope: ScopeT | None = None, raise_json_decode_error: bool = False
    ) -> str | None:
        try:
            message_dict = json.loads(message)
        except json.JSONDecodeError as e:
            if raise_json_decode_error:
                raise
            response = to_dict(json_rpc.build_error_message(types.PARSE_ERROR, {}, e))
        else:
            response = await self.handle(message_dict, scope)

        return json.dumps(response) if response is not None else None

    async def _handle_rpc_msg(self, rpc_msg: types.JSONRPCMessage) -> types.JSONRPCMessage | None:
        msg_root = rpc_msg.root

        # --- Handle request ---
        if isinstance(msg_root, types.JSONRPCRequest):
            client_request = types.ClientRequest.model_validate(to_dict(msg_root))

            logger.info(f"Handling request: {client_request}")
            response = await self._handle_client_request(client_request)
            logger.info(f"Handled request: {client_request}")

            if response is None:
                # Request was cancelled - Do nothing
                return None

            logger.info(f"Returning response: {response}")
            return json_rpc.build_response_message(msg_root.id, response)

        # --- Handle notification ---
        elif isinstance(msg_root, types.JSONRPCNotification):
            # TODO: Add full support for client notification - This just implements the handler.
            client_notification = types.ClientNotification.model_validate(to_dict(msg_root))

            logger.info(f"Handling notification: {client_notification}")
            await self._handle_client_notification(client_notification)
            logger.info(f"Handled notification: {client_notification}")
            return None
        else:
            raise UnsupportedRPCMessageType("Message to MCP server must be a request or notification")

    async def _handle_client_request(self, request: types.ClientRequest) -> types.ServerResult:
        if handler := self._core.request_handlers.get(type(request.root)):  # type: ignore
            logger.debug("Dispatching request of type %s", type(request.root).__name__)
            return await handler(request.root)
        else:
            raise McpError(types.ErrorData(code=types.METHOD_NOT_FOUND, message="Method not found"))

    async def _handle_client_notification(self, notification: types.ClientNotification):
        notify = notification.root
        if handler := self._core.notification_handlers.get(type(notify)):  # type: ignore
            logger.debug("Dispatching notification of type %s", type(notify).__name__)

            try:
                await handler(notify)
            except Exception:
                logger.exception("Uncaught exception in notification handler")

    async def _initialize_handler(self, req: types.InitializeRequest) -> types.ServerResult:
        client_protocol_version = req.params.protocolVersion
        server_protocol_version = (
            client_protocol_version
            if client_protocol_version in version.SUPPORTED_PROTOCOL_VERSIONS
            else types.LATEST_PROTOCOL_VERSION
        )

        init_options = self._core.create_initialization_options(
            notification_options=self._notification_options,
        )

        init_result = types.InitializeResult(
            protocolVersion=server_protocol_version,
            capabilities=init_options.capabilities,
            serverInfo=types.Implementation(
                name=init_options.server_name,
                version=init_options.server_version,
            ),
            instructions=init_options.instructions,
        )

        return types.ServerResult(init_result)



import logging
from contextlib import contextmanager
from contextvars import ContextVar, Token
from dataclasses import dataclass
from typing import Generic, TypeVar

from mcp.types import JSONRPCMessage

from minimcp.server.exceptions import ContextError

logger = logging.getLogger(__name__)


ScopeT = TypeVar("ScopeT", bound=object)


@dataclass
class Context(Generic[ScopeT]):
    message: JSONRPCMessage
    scope: ScopeT | None = None


class ContextManager(Generic[ScopeT]):
    _ctx: ContextVar[Context[ScopeT]] = ContextVar("ctx")

    @contextmanager
    def active(self, message: JSONRPCMessage, scope: ScopeT | None = None):
        # Build context
        context = Context(message=message, scope=scope)

        # Set context
        token: Token = self._ctx.set(context)
        try:
            yield
        finally:
            # Clear context
            self._ctx.reset(token)

    def _get(self) -> Context[ScopeT]:
        try:
            return self._ctx.get()
        except LookupError:
            err = ContextError("No Context: Called get outside of an active context")
            logger.error(err)
            raise err

    def get_message(self) -> JSONRPCMessage:
        return self._get().message

    def get_scope(self) -> ScopeT:
        scope = self._get().scope
        if scope is None:
            raise ContextError("No Scope: Scope is not set in the context")
        return scope



import builtins
import inspect
from collections.abc import Callable
from functools import partial
from typing import Any

import mcp.types as types
from mcp.server.fastmcp.utilities.func_metadata import FuncMetadata, func_metadata
from mcp.server.lowlevel.server import Server
from typing_extensions import TypedDict, Unpack


class ToolDetails(TypedDict, total=False):
    name: str | None
    description: str | None
    annotations: types.ToolAnnotations | None
    meta: dict[str, Any] | None


class ToolManager:
    """
    Manages tool definitions and handlers.
    """

    _tools: dict[str, tuple[types.Tool, types.AnyFunction, FuncMetadata]]

    def __init__(self, core: Server):
        self._tools = {}
        self._hook_core(core)

    def __call__(self, **kwargs: Unpack[ToolDetails]) -> Callable[[Callable], types.Tool]:
        """
        Decorator to add a tool to the MCP tool manager.
        """
        return partial(self.add, **kwargs)

    def add(self, func: types.AnyFunction, **kwargs: Unpack[ToolDetails]) -> types.Tool:
        """
        Add a tool to the MCP tool manager.
        """

        tool_name = kwargs.get("name", func.__name__)

        if not tool_name:
            raise ValueError("Tool name is required")

        if tool_name in self._tools:
            raise ValueError(f"Tool {tool_name} already registered")

        func_meta = func_metadata(func)
        parameters = func_meta.arg_model.model_json_schema(by_alias=True)

        tool = types.Tool(
            name=tool_name,
            description=kwargs.get("description", func.__doc__),
            inputSchema=parameters,
            outputSchema=func_meta.output_schema,
            annotations=kwargs.get("annotations", None),
            _meta=kwargs.get("meta", None),
        )

        self._tools[tool_name] = (tool, func, func_meta)

        return tool

    def remove(self, name: str) -> types.Tool:
        """
        Remove a tool from the MCP tool manager.
        """
        if name not in self._tools:
            raise ValueError(f"Tool {name} not found")

        return self._tools.pop(name)[0]

    def _hook_core(self, core: Server):
        core.list_tools()(self._async_list)
        # Validation done by func_meta in call. Hence passing validate_input=False
        # TODO: Ensure both the validations are similar
        core.call_tool(validate_input=False)(self.call)

    async def _async_list(self) -> builtins.list[types.Tool]:
        return self.list()

    def list(self) -> builtins.list[types.Tool]:
        return [tool[0] for tool in self._tools.values()]

    async def call(self, name: str, args: dict[str, Any]) -> Any:
        """
        Call a tool - Can be called from anywhere.
        """

        if name not in self._tools:
            raise ValueError(f"Tool {name} not found")

        _, handler, func_meta = self._tools[name]

        parsed_args = func_meta.pre_parse_json(args)
        validated_args = func_meta.arg_model.model_validate(parsed_args)
        result = handler(**validated_args.model_dump_one_level())

        if inspect.isawaitable(result):
            result = await result

        return func_meta.convert_result(result)



from typing import Any

from mcp.types import ErrorData, JSONRPCError, JSONRPCMessage, JSONRPCResponse, ServerResult

from minimcp.server.utils import to_dict

JSON_RPC_VERSION = "2.0"


def build_response_message(message_id: str | int, response: ServerResult) -> JSONRPCMessage:
    return JSONRPCMessage(JSONRPCResponse(jsonrpc=JSON_RPC_VERSION, id=message_id, result=to_dict(response)))


def build_error_message(
    error_code: int, rpc_message: dict[str, Any] | None, error: BaseException, error_data: ErrorData | None = None
) -> JSONRPCMessage:
    message_id = rpc_message.get("id", "") if isinstance(rpc_message, dict) else ""
    error_message = f"{error.__class__.__name__}: {error}"
    error_data = error_data or ErrorData(code=error_code, message=error_message, data=None)

    return JSONRPCMessage(JSONRPCError(jsonrpc=JSON_RPC_VERSION, id=message_id, error=error_data))



"""
A simple MCP server for mathematical operations.
"""

from pydantic import Field

from minimcp.server import MiniMCP

math_mcp = MiniMCP(name="MathServer", version="0.1.0")


@math_mcp.tool(description="Add two numbers")
def add(
    a: float = Field(description="The first float number"), b: float = Field(description="The second float number")
) -> float:
    return a + b


@math_mcp.tool(description="Subtract two numbers")
def subtract(
    a: float = Field(description="The first float number"), b: float = Field(description="The second float number")
) -> float:
    return a - b


@math_mcp.tool(description="Multiply two numbers")
def multiply(
    a: float = Field(description="The first float number"), b: float = Field(description="The second float number")
) -> float:
    return a * b


@math_mcp.tool(description="Divide two numbers")
def divide(
    a: float = Field(description="The first float number"), b: float = Field(description="The second float number")
) -> float:
    return a / b
